/*
# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
  This application demonstrates the use of the library libModBus for
  the C programming language on how to communicate with an Analog One
  device via a serial line using ModBus RTU. The program will first
  set the analog output 0 to 5 volts then read the analog input 0, then
  sets the analog output 0 to 10 volts and then reads the analog input 0
  again. Each write and read action will be printed to the console so
  the user can follow the applications tasks. For this app to work
  correctly, a wire needs to be connected between AO0 and AI0.
  
  For this application demonstration we assume that it is compiled and
  run on a PiXtend V2 -L- which has an RS485 serial port.
  We will also be using the ttyAMA0 Raspberry Pi serial device, make sure
  it is available to use. If you are not sure please see the Software manual
  for the PiXtend eIO devices and have a look at the Basic Knowledge section.
  
  Notice: Only the ttyAMA0 serial device supports the parity setting,
          the ttyS0 does not.
          
  To compile the app try:
  : sudo apt update
  : sudo apt install libmodbus-dev
  : gcc -I /usr/include/modbus PiXtend_eIO_Analog_One_ModBus_Demo_AO0_to_AI0.c -o PiXtend_eIO_Analog_One_ModBus_Demo_AO0_to_AI0 -L/usr/lib/modbus -lmodbus -lwiringPi

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <wiringPi.h>
#include <modbus.h>

//By default the Analog One device has
//a ModBus RTU address of "3" (factory setting)
#define SLAVE_ID  3
//How often do we want to loop the tasks
#define MAX_LOOPS 100

//Remember if we did the initalization already
static uint8_t byInitFlag = 0;
struct timespec ts;

int setup(void)
{	
    //Set wiringPi GPIO 1 which is also BCM GPIO 18
    int pin_RS485_enable = 1;
    
	if(byInitFlag < 1)
	{
        //Setup wiringPi functions
		wiringPiSetup();
		byInitFlag = 1;
	}
	
    //Enable the RS485 port on the PiXtend V2 -L- so we can
    //talk to the PiXtend eIO device "Analog One"
	pinMode(pin_RS485_enable, OUTPUT);
	digitalWrite(pin_RS485_enable, 1);

    //Setup time struct for 5 second sleep in main program
    ts.tv_sec = 5;
    ts.tv_nsec = 5000000;    

	return 0;
}

/*
  Main program

  We use the "modbus_write_register" and "modbus_read_input_registers" functions
  to write and read data from the Analog One device. Using these functions
  allows us to set the analog output 0 to 5 volts and then to 10 volts and so
  on. After a short wait we read the voltage back via the analog input 0 to see
  if everything worked out. Remember to put a cable in analog output 0 and
  analog input 0 which connects these terminals together to make this app work.
  
  Feel free to adapt this program to your need or just to play around.
*/
int main(void)
{
    //ModBus context variable
    modbus_t *ctx;
    //Holds the return value of a ModBus function 
    int retval;
    //Holds a return value
    int res;
    //Count the number of unsuccessful transfers
    int modbus_error_cnt;
    //Simple counter variable
    int i;
    //Task counter variable
    int task;
    //ModBus RTU address of AO0 Holding Register
    uint16_t ao0_register_address = 0;
    //ModBus RTU address of AI0 Input Register
    uint16_t ai0_register_address = 0;
    //AO value for 5 volts, this equals about 2047 DAC units
    uint16_t ao0_5_volts_value = 2047;
    //AO value for 10 volts, this equals about 4095 DAC units
    uint16_t ao0_10_volts_value = 4095;
    //Pointer to store the returned Input Register array
    uint16_t *pinp_registers = NULL;

    //Perform needed setup steps at the beginning of the program.
    //We need to setup the RS485 port.
    setup();

    /* RTU */
    //Use serial port ttyAMA0 with Analog One facotry settings of:
    //19200 baud, 8 data bits, 1 stop bit and Even parity
    ctx = modbus_new_rtu("/dev/ttyAMA0", 19200, 'E', 8, 1);
    //Set the Slave ID of the Analog One device also called 
    //a Bus Address or Device ID
    modbus_set_slave(ctx, SLAVE_ID);
    //For normal operation we do not need to enable the debug option.
    //If you are havig trouble connecting to the ModBus device, you can
    //uncomment the next line, start the program again and look at the
    //debug output of the ModBus RTU library, to see what is going wrong.
    //modbus_set_debug(ctx, TRUE);

    //Check if the connection was successful
    if (modbus_connect(ctx) == -1) {
        fprintf(stderr, "Connection failed: %s\n",
                modbus_strerror(errno));
        modbus_free(ctx);
        return -1;
    }

    // Allocate and initialize the memory space
    // We need space for 1 Input Register
    pinp_registers = (uint16_t *) malloc(1 * sizeof(uint16_t));
    memset(pinp_registers, 0, 1 * sizeof(uint16_t));

    //Pre-set runtime variables
    i = modbus_error_cnt = task = 0;

    //Loop as long as i is smaller then MAX_LOOPS
    while (i++ < MAX_LOOPS) 
    {
        //Task 0 - Set the Analog Output 0 to 5 volts
        if (task == 0)
        {
            printf("Write Analog Output 0 to 5 volts (2047 units)\n");
            //Write the holding register 0 of the Analog One device
            retval = modbus_write_register(ctx, ao0_register_address, ao0_5_volts_value);
            //Check if the transfer was successful or not
            if (retval != 1) {
                //There was an error writing the register to the slave
                printf("ERROR modbus_write_register (%d)\n", retval);
                modbus_error_cnt++;
            }
            else
            {
                //Everything went fine, data was written to register
                printf("OK\n", retval);
            }

            task++;
        }
        //Task 1 - Read the Input Register 0, which is the Analog Input 0
        //of the Analog One device
        else if (task == 1)
        {
            printf("Read Analog Input 0\n");
            //Read Input Register 0 which has the value of Analog Input 0
            retval = modbus_read_input_registers(ctx, ai0_register_address, 1, pinp_registers);
            //Check if the transfer was successful or not
            if (retval != 1) {
                //There was an error writing the register to the slave
                printf("ERROR modbus_read_input_registers (%d)\n", retval);
                modbus_error_cnt++;
            }
            else
            {
                //Everything went fine, data was written to register
                printf("OK - AI0 = %d (should be around 511 units)\n\n", pinp_registers[0]);
            }

            task++;            
        }
        //Task 2 - Set the Analog Output 0 to 10 volts
        else if (task == 2)
        {
            printf("Write Analog Output 0 to 10 volts (4095 units)\n");
            //Write the holding register 0 of the Analog One device
            retval = modbus_write_register(ctx, ao0_register_address, ao0_10_volts_value);
            //Check if the transfer was successful or not
            if (retval != 1) {
                //There was an error writing the register to the slave
                printf("ERROR modbus_write_register (%d)\n", retval);
                modbus_error_cnt++;
            }
            else
            {
                //Everything went fine, data was written to register
                printf("OK\n", retval);
            }

            task++;            
        }
        //Task 3 - Read the Input Register 0, which is the Analog Input 0
        //of the Analog One device and print out its value
        else if (task == 3)
        {
            printf("Read Analog Input 0\n");
            //Read Input Register 0 which has the value of Analog Input 0
            retval = modbus_read_input_registers(ctx, ai0_register_address, 1, pinp_registers);
            //Check if the transfer was successful or not
            if (retval != 1) {
                //There was an error writing the register to the slave
                printf("ERROR modbus_read_input_registers (%d)\n\n", retval);
                modbus_error_cnt++;
            }
            else
            {
                //Everything went fine, data was written to register
                printf("OK - AI0 = %d  (should be around 1023 units)\n", pinp_registers[0]);
            }
            //Do the first task again and begin a new cycle
            task = 0;
            printf("\n");
            printf("----------------------------------------\n");
            printf("\n"); 
        }    
        
        //After each transfer sleep 5 seconds
        do{
         res = nanosleep(&ts, &ts);
        }while (res && errno == EINTR);
    }

    // Free the memory
    free(pinp_registers);

    // Close the connection
    modbus_close(ctx);
    modbus_free(ctx);

    //End the program
    return 0;
}



