/*
# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
  This demonstration app shows how to communicate with the PiXtend eIO
  Analog One device using the PiXtend eIO Protocol (ASCII based).
  Using an array of strings with pre defined commands, the analog output 0
  is set to 5 volts and then to 10 volts, alternating every few seconds.
  After each change the value of the analog input 0 is read. This way this
  example shows how to send a command to the device and then read back the
  reply and process the received information.
  
  To compile:
  gcc -Wall -o "PiXtend_eIO_Demo_Analog_Output-Input" "PiXtend_eIO_Demo_Analog_Output-Input.c" -lpixtend -lwiringPi
  
  To run:
  sudo ./PiXtend_eIO_Demo_Analog_Output-Input
*/

#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <pixtend.h>
#include <termios.h>
#include <unistd.h>

// How often do we want the DO leds to turn on and off
#define MAX_LOOPS 100
// Remember if we did the initalization already
static uint8_t byInitFlag = 0;
// Variable for serial port file descriptor
static int fd;
// Options structure with serial port options, this is
// needed to be able to configure the serial port to use
// EVEN paraity and to activate a timeout for blocking
// read operations.
struct termios options;
// Pre defined list with PiXtend eIO commands for Analog One
// Set analog output 0
const char *  msgout[] = {"#003:006:00:2047*",
                          "#003:006:00:0000*",
                          "#003:006:00:4095*",
                          "#003:006:00:0000*"};
// Pre defined list with PiXtend eIO commands for Analog One
// Read analog input 0
const char *  msginp[] = {"#003:005:00:0000*",
                          "#003:005:00:0000*",
                          "#003:005:00:0000*",
                          "#003:005:00:0000*"};
// Store the reply from the Analog One device
char ai0_reply[17];
// This array will hold only the analog input 0 value
char ai0_value[5];

// Signal handler if the user presses Ctrl+C to abort the
// program, so we have to close the serial port by any means.
void signal_callback_handler(int signum)
{
	printf("\n\nCaught signal %d\n",signum);
	serialFlush(fd);
	printf("Flushed serial device\n");
	serialClose(fd);
	printf("Closed serial device\n");
	exit(signum);
}

// Setup the PiXtend V2 -L- and activate the RS485 port
int setup(void)
{	
    // Set wiringPi GPIO 1 which is also BCM GPIO 18
    int pin_RS485_enable = 1;
    
	if(byInitFlag < 1)
	{
        // Setup wiringPi functions
		wiringPiSetup();
		byInitFlag = 1;
	}
	
    // Enable the RS485 port on the PiXtend V2 -L- so we can
    // talk to the PiXtend eIO device "Analog One"
	pinMode(pin_RS485_enable, OUTPUT);
	digitalWrite(pin_RS485_enable, 1);    

	return 0;
}

/*
  Main program

  We use the serial port to send commands to an Analog One device to change
  the analog output 0 and read back the value from analog input 0.
  The commands are send out in a cyclic manner, but the loop will terminate
  automatically after 100 loops.
*/
int main(void)
{
    // We need to use the ttyAMA0 device so we can use paraity
    char strDevice[] = "/dev/ttyAMA0";
    // By default the Analog One device uses 19200 baud (factory setting)
    int  baudRate = 19200;
    // Counter variables for the main program
    int i, j;
    int chr, nRet, nRetTemp;
    
    // Do the PiXtend V2 -L- setup and activate RS485
    setup();
    // Open the serial port and store the file descriptor
    fd = serialOpen(strDevice,baudRate);
    // Check if we have a valid file descriptor and that serial port
    // could be opned correctly
    if(fd==-1)
    {
        // There was an error, serial port could not be opned.
        // Not enough privliges? Try running with sudo or as root
        printf("Error: Couldn't open serial device\n");
    }
    else
    {
        // The serial device is open an we can talk now to the
        // Analog One device and others if we want to
        printf("Opened serial device [fd=%d]\n",fd);
        // Get the currently active options from the serial port 
        tcgetattr (fd, &options) ;
        // Enable Parity - EVEN by default
        options.c_cflag |= PARENB;
        // Set timeout values and make the timeout function active
        options.c_cc[VMIN]  = 1;
        // 0.5 seconds read timeout 
        options.c_cc[VTIME] = 5;            
        // Set new options 
        if (tcsetattr(fd, TCSANOW, &options) != 0) {
            printf("Error from tcsetattr: %s\n", strerror(errno));
            return -1;
        }  
        // Activate the signal handler in case we press Ctrl+C while
        // the program is running
        signal(SIGINT, signal_callback_handler);
        // Flush the console output
        fflush(stdout);
        
        // Loop as long as i is smaller then MAX_LOOPS
        while (i++ < MAX_LOOPS) 
        {
            // Print some information how many loops we have done already
            printf ("\n");
            printf ("----------------------------------------\n");
            printf ("Doing loop number: %d/%d\n", i, MAX_LOOPS);
            printf ("----------------------------------------\n");
            // Loop through all 4 commands for writing and reading from
            // an Analog One device, which we pre defined in the beginning.
            for (j = 0; j < 4; j++) 
            {
                // Print the current command string to console
                printf ("Write AO0 - %s\n", msgout[j]);
                // Write the string to the serial port and send it
                // to the Analog One device
                serialPuts(fd, msgout[j]);

                // Uncomment this section if you also want to see the
                // return string / reply of the Analog One device in
                // case further actions or validation is needed. This
                // section is meant as an example how reading from the
                // serial port can be done. Adapt as needed or just play
                // around and see what happens.
                // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                /*
                
                // Wait for the 17 byte transfer to complete, each byte
                // takes about 450 us.
                usleep (17 * 450);
                // Wait 10 ms for the Device to process the command before
                // sendig it back
                usleep (10000);
                // Wait again for the 17 byte of the reply to arrive at the
                // PiXtend V2 -L- serial port and to be stored in the OSs buffer
                usleep (17 * 450);
                // Read the bytes one by one from the buffer until it is empty
                nRetTemp = 0;
                printf ("Reply:   ");
                do
				{
                    // Check if data (bytes) is available
					nRet = serialDataAvail(fd);
					if (nRet>=0)
					{
                        // Get / read one character from the serial port buffer
						chr = serialGetchar(fd);
                        // Did we get something?
						if(chr!=-1)
						{
                            // Print the returned character to console
							printf("%c",chr);
                            // Show the data on the console right away
							fflush(stdout);
						}
                        // Count to see if we reached 16 bytes,
                        // if so we can stop and leave the loop
                        nRetTemp = nRetTemp + 1;
					}
					else
					{
                        // There was an error getting data from the serial port
						printf("Error: serialDataAvail=-1\n");
                        // Clear out the serial port
						serialFlush(fd);
						printf("Flushed serial device\n");
                        // Close the serial port
						serialClose(fd);
						printf("Closed serial device\n");
                        //Leave the program
						return 0;
					}
                // Loop until we have received atleast 17 bytes, this is the max length
                // of the PiXtend eIO protocol for the Analog One. For a real application
                // there should probably also be a timer supervision, if there is no
                // complete response for example within 1 second then abort
				}while (nRetTemp < 17);
                printf ("\n");
                printf ("----------------------------------------\n");
                
                */
                // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                
                //Sleep 0.25 seconds after this command
                usleep (250000);

                // Clear out the serial port buffer for new data
				serialFlush(fd);
                        
                // Print the current command string to console
                printf ("Read AI0  - %s\n", msginp[j]);
                // Write the string to the serial port and send it
                // to the Analog One device
                serialPuts(fd, msginp[j]);
                // Wait for the 17 byte transfer to complete, each byte
                // takes about 450 us.
                usleep (17 * 450);
                // Wait 10 ms for the Device to process the command before
                // sendig it back
                usleep (10000);
                // Wait again for the 17 byte of the reply to arrive at the
                // PiXtend V2 -L- serial port and to be stored in the OSs buffer
                usleep (17 * 450);
                // Read the bytes one by one from the buffer until it is empty
                nRetTemp = 0;
                printf ("Reply     - ");
                do
				{
                    // Check if data (bytes) is available
					nRet = serialDataAvail(fd);
					if (nRet>=0)
					{
                        // Get / read one character from the serial port buffer
						chr = serialGetchar(fd);
                        // Did we get something?
						if(chr!=-1)
						{
                            // Print the returned character to console
							printf("%c",chr);
                            // Save the received char for later use
                            ai0_reply[nRetTemp] = chr;
                            // Show the data on the console right away
							fflush(stdout);
						}
                        // Count to see if we reached 17 bytes,
                        // if so we can stop and leave the loop
                        nRetTemp = nRetTemp + 1;
					}
					else
					{
                        // There was an error getting data from the serial port
						printf("Error: serialDataAvail=-1\n");
                        // Clear out the serial port
						serialFlush(fd);
						printf("Flushed serial device\n");
                        // Close the serial port
						serialClose(fd);
						printf("Closed serial device\n");
                        //Leave the program
						return 0;
					}
                // Loop until we have received atleast 17 bytes, this is the max length
                // of the PiXtend eIO protocol for the Analog One. For a real application
                // there should probably also be a timer supervision, if there is no
                // complete response for example within 1 second then abort
				}while (nRetTemp < 17);
                printf ("\n");
                // Convert the value from Analog Input 0 from string to integer and
                // and print it out
                ai0_value[0] = ai0_reply[12];
                ai0_value[1] = ai0_reply[13];
                ai0_value[2] = ai0_reply[14];
                ai0_value[3] = ai0_reply[15];
                ai0_value[4] = 0;
                int result = atoi(ai0_value);
                printf ("Value AI0 - %d\n", result);
                printf ("----------------------------------------\n");
                
                //Sleep 2 seconds after this command
                usleep (1000000);
                usleep (1000000);
            }
        }
        // Once the loop is done or the user presses Ctrl+C, close the serial port and leave
        serialClose(fd);
        printf("Closed serial device\n");        
    }
    return 0;
}