/*
# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
  This application demonstrates the use of the library libModBus for
  the C programming language on how to communicate with a Digital One
  device and let the first digital output blink.

  For this application demonstration we assume that it is compiled and
  run on a PiXtend V2 -L- which has an RS485 serial port.
  We will also be using the ttyAMA0 Raspberry Pi serial device, make sure
  it is available to use. If you are not sure please see the Software manual
  for the PiXtend eIO devices and have a look at the Basic Knowledge section.
  
  Notice: Only the ttyAMA0 serial device supports the parity setting,
          the ttyS0 does not.
          
  To compile the app try:
  : sudo apt update
  : sudo apt install libmodbus-dev
  : gcc -I /usr/include/modbus PiXtend_eIO_Digital_One_ModBus_Demo_Blinky.c -o PiXtend_eIO_Digital_One_ModBus_Demo_Blinky -L/usr/lib/modbus -lmodbus -lwiringPi

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <wiringPi.h>
#include <modbus.h>

//By default the Digital One device has
//a ModBus RTU address of "1" (factory setting)
#define SLAVE_ID  1
//How often do we want the DO0 led to blink
#define MAX_LOOPS 100

//Remember if we did the initalization already
static uint8_t byInitFlag = 0;
struct timespec ts;

int setup(void)
{	
    //Set wiringPi GPIO 1 which is also BCM GPIO 18
    int pin_RS485_enable = 1;
    
	if(byInitFlag < 1)
	{
        //Setup wiringPi functions
		wiringPiSetup();
		byInitFlag = 1;
	}
	
    //Enable the RS485 port on the PiXtend V2 -L- so we can
    //talk to the PiXtend eIO device "Digital One"
	pinMode(pin_RS485_enable, OUTPUT);
	digitalWrite(pin_RS485_enable, 1);

    //Setup time struct for 1 second sleep in main program
    ts.tv_sec = 1;
    ts.tv_nsec = 1000000;    

	return 0;
}

/*
  Main program

  We use the "modbus_write_bits" function (FC15) to write multiple Coils
  atonce. In this demo we will only let the digital output 0 blink for
  a few times and then end the program automatically. Feel free to adapt
  this program to your need or just to play around.
*/
int main(void)
{
    //ModBus context variable
    modbus_t *ctx;
    //Holds the return value of modbus_write_bits function 
    int retval;
    //Holds a return value
    int res;
    //Count the number of unsuccessful transfers
    int modbus_error_cnt;
    //Simple counter variable
    int i;
    //ModBus RTU start address of the first Coil
    int coils_start_address;
    //Number of available Coils (bits) in one byte
    int number_of_bits;
    //Memory area for the Coils
    uint8_t *modbus_coils;
    //Variable to help toggle the digital output 0 LED on/off
    int coil_on;

    //Perform needed setup steps at the beginning of the program.
    //We need to setup the RS485 port.
    setup();

    /* RTU */
    //Use serial port ttyAMA0 with Digital One facotry settings of:
    // 19200 baud, 8 data bits, 1 stop bit and Even parity
    ctx = modbus_new_rtu("/dev/ttyAMA0", 19200, 'E', 8, 1);
    //Set the Slave ID of the Digital One device also called 
    //a Bus Address or Device ID
    modbus_set_slave(ctx, SLAVE_ID);
    //For normal operation we do not need to enable the debug option.
    //If you are havig trouble connecting to the ModBus device, you can
    //uncomment the next line, start the program again and look at the
    //debug output of the ModBus RTU library, to see what is going wrong.
    //modbus_set_debug(ctx, TRUE);

    //Check if the connection was successful
    if (modbus_connect(ctx) == -1) {
        fprintf(stderr, "Connection failed: %s\n",
                modbus_strerror(errno));
        modbus_free(ctx);
        return -1;
    }

    // Allocate and initialize the memory space
    //We use 8 bits per byte, meaning we have 8 Coils which
    //also corresponds to the number of digital outputs the
    //Digital One device has.
    number_of_bits = 8;

    modbus_coils = (uint8_t *) malloc(number_of_bits * sizeof(uint8_t));
    memset(modbus_coils, 0, number_of_bits * sizeof(uint8_t));

    //Pre-set runtime variables
    i = modbus_error_cnt = 0;
    coils_start_address = 0;
    coil_on = 1;
    
    //Loop as long as i is smaller then MAX_LOOPS
    while (i++ < MAX_LOOPS) 
    {
        //Check if DO0 should go ON
        if (coil_on == 1)
        {
           //Set DO0 to ON
           modbus_coils[0] = 1;
           coil_on = 0;
        }
        else
        {
            //Turn DO0 OFF
            modbus_coils[0] = 0;
            coil_on = 1;
        }
        
        //Write the data to the ModBus RTU device, transfer all Coils atonce
        retval = modbus_write_bits(ctx, coils_start_address, number_of_bits, modbus_coils);
        //Check if the transfer was successful or not
        if (retval != 8) {
            //There was an error writing the Coils (bits) to the slave
            printf("ERROR modbus_write_bits (%d)\n", retval);
            modbus_error_cnt++;
        }
        else
        {
            //Everything went fine, all Coild (bits) were transfered
            printf("OK\n", retval);
        }
        
        //After each transfer sleep 1 second
        do{
         res = nanosleep(&ts, &ts);
        }while (res && errno == EINTR);
    }

    // Free the memory
    free(modbus_coils);

    // Close the connection
    modbus_close(ctx);
    modbus_free(ctx);

    //End the program
    return 0;
}



