/*
# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
  This demonstration app shows how to communicate with the PiXtend eIO
  Digital One device using the PiXtend eIO Protocol (ASCII based).
  Using an array of strings with predefined commands, each digital output
  is turned on and off after each other to simulate something we call a
  running light, as the light moves from left to right and back again.
  
  To compile:
  gcc -Wall -o "PiXtend_eIO_Demo_Running_Lights" "PiXtend_eIO_Demo_Running_Lights.c" -lpixtend -lwiringPi
  
  To run:
  sudo ./PiXtend_eIO_Demo_Running_Lights
*/

#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <pixtend.h>
#include <termios.h>
#include <unistd.h>

// How often do we want the DO leds to turn on and off
#define MAX_LOOPS 100
// Remember if we did the initalization already
static uint8_t byInitFlag = 0;
// Variable for serial port file descriptor
static int fd;
// Options structure with serial port options, this is
// needed to be able to configure the serial port to use
// EVEN paraity and to activate a timeout for blocking
// read operations.
struct termios options;
// Pre defined list with PiXtend eIO commands for the
// Digital One device. Each line turns 1 LED of a
// digital output on and all others off. This way
// it looks like the light is moving from left to right
// and back again.
const char *  message[] = {"#001:004:01:001*",
                           "#001:004:01:002*",
                           "#001:004:01:004*",
                           "#001:004:01:008*", 
                           "#001:004:01:016*", 
                           "#001:004:01:032*", 
                           "#001:004:01:064*",
                           "#001:004:01:128*",
                           "#001:004:01:128*",
                           "#001:004:01:064*",
                           "#001:004:01:032*",
                           "#001:004:01:016*",
                           "#001:004:01:008*",
                           "#001:004:01:004*",
                           "#001:004:01:002*",
                           "#001:004:01:001*"};

// Signal handler if the user presses Ctrl+C to abort the
// program, so we have to close the serial port by any means.
void signal_callback_handler(int signum)
{
	printf("\n\nCaught signal %d\n",signum);
	serialFlush(fd);
	printf("Flushed serial device\n");
	serialClose(fd);
	printf("Closed serial device\n");
	exit(signum);
}

// Setup the PiXtend V2 -L- and activate the RS485 port
int setup(void)
{	
    // Set wiringPi GPIO 1 which is also BCM GPIO 18
    int pin_RS485_enable = 1;
    
	if(byInitFlag < 1)
	{
        // Setup wiringPi functions
		wiringPiSetup();
		byInitFlag = 1;
	}
	
    // Enable the RS485 port on the PiXtend V2 -L- so we can
    // talk to the PiXtend eIO device "Digital One"
	pinMode(pin_RS485_enable, OUTPUT);
	digitalWrite(pin_RS485_enable, 1);    

	return 0;
}

/*
  Main program

  We use the serial port to send a Digital One device with bus ID 1 a
  series of pre defined commands which we take from an array variable.
  The commands turn each digital output on and off one after another,
  which also turn on the LED of each output and the idea is to create
  the illusion that the LED light moves from left to right and bach
  again. It runs from left to right from right to left :-)
  
  The program ends automatically after 100 complete loops where every
  inner loop sends up to 16 commands to the Digital One device.
  After each command there is a 250 ms (0.25 sec) pause. If the pause
  is shortened, the light moves faster, however, do not delete the pause!
*/
int main(void)
{
    // We need to use the ttyAMA0 device so we can use paraity
    char strDevice[] = "/dev/ttyAMA0";
    // By default the Digital One device uses 19200 baud (factory setting)
    int  baudRate = 19200;
    // Counter variables for the main program
    int i, j;
    // Uncomment this declaration if you also uncommend the middle part of the
    // main program to also see the replies of the Digital One device.
    //int chr, nRet, nRetTemp;
    
    // Do the PiXtend V2 -L- setup and activate RS485
    setup();
    // Open the serial port and store the file descriptor
    fd = serialOpen(strDevice,baudRate);
    // Check if we have a valid file descriptor and that serial port
    // could be opned correctly
    if(fd==-1)
    {
        // There was an error, serial port could not be opned.
        // Not enough privliges? Try running with sudo or as root
        printf("Error: Couldn't open serial device\n");
    }
    else
    {
        // The serial device is open an we can talk now to the
        // Digital One device and others if we want to
        printf("Opened serial device [fd=%d]\n",fd);
        // Get the currently active options from the serial port 
        tcgetattr (fd, &options) ;
        // Enable Parity - EVEN by default
        options.c_cflag |= PARENB;
        // Set timeout values and make the timeout function active
        options.c_cc[VMIN]  = 1;
        // 0.5 seconds read timeout 
        options.c_cc[VTIME] = 5;            
        // Set new options 
        if (tcsetattr(fd, TCSANOW, &options) != 0) {
            printf("Error from tcsetattr: %s\n", strerror(errno));
            return -1;
        }  
        // Activate the signal handler in case we press Ctrl+C while
        // the program is running
        signal(SIGINT, signal_callback_handler);
        // Flush the console output
        fflush(stdout);
        
        //Loop as long as i is smaller then MAX_LOOPS
        while (i++ < MAX_LOOPS) 
        {
            // Print some information how many loops we have done already
            printf ("\n");
            printf ("----------------------------------------\n");
            printf ("Doing loop number: %d/%d\n", i, MAX_LOOPS);
            printf ("----------------------------------------\n");
            printf ("\n");
            // Loop through all 16 array elements and send the strings
            // to the Digital One device
            for (j = 0; j < 16; j++) 
            {
                // Print the current command string to console
                printf ("Send - %d: %s\n", j, message[j]);
                // Write the string to the serial port and send it
                // to the Digital One device
                serialPuts(fd, message[j]);

                // Uncomment this section if you also want to see the
                // return string / reply of the Digital One device in
                // case further actions or validation is needed. This
                // section is meant as an example how reading from the
                // serial port can be done. Adapt as needed or just play
                // around and see what happens.
                // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                /*
                
                // Wait for the 16 byte transfer to complete, each byte
                // takes about 450 us.
                usleep (16 * 450);
                // Wait 10 ms for the Device to process the command before
                // sendig it back
                usleep (10000);
                // Wait again for the 16 byte of the reply to arrive at the
                // PiXtend V2 -L- serial port and to be stored in the OSs buffer
                usleep (16 * 450);
                // Read the bytes one by one from the buffer until it is empty
                nRetTemp = 0;
                printf ("Reply:   ");
                do
				{
                    // Check if data (bytes) is available
					nRet = serialDataAvail(fd);
					if (nRet>=0)
					{
                        // Get / read one character from the serial port buffer
						chr = serialGetchar(fd);
                        // Did we get something?
						if(chr!=-1)
						{
                            // Print the returned character to console
							printf("%c",chr);
                            // Show the data on the console right away
							fflush(stdout);
						}
                        // Count to see if we reached 16 bytes,
                        // if so we can stop and leave the loop
                        nRetTemp = nRetTemp + 1;
					}
					else
					{
                        // There was an error getting data from the serial port
						printf("Error: serialDataAvail=-1\n");
                        // Clear out the serial port
						serialFlush(fd);
						printf("Flushed serial device\n");
                        // Close the serial port
						serialClose(fd);
						printf("Closed serial device\n");
                        //Leave the program
						return 0;
					}
                // Loop until we have received atleast 16 bytes, this is the max length
                // of the PiXtend eIO protocol for the Digital One. For a real application
                // there should probably also be a timer supervision, if there is no
                // complete response for example within 1 second then abort
				}while (nRetTemp < 16);
                printf ("\n");
                printf ("----------------------------------------\n");
                
                */
                // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                
                //Sleep 0.25 seconds after each command
                usleep (250000);
                
            }
        }
        // Once the loop is done or the user presses Ctrl+C, close the serial port and leave
        serialClose(fd);
        printf("Closed serial device\n");        
    }
    return 0;
}