﻿using System;
using System.IO.Ports;
using System.Threading;

/*
# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Instructions
 * 
 * Set the COM port of the RS485 dongle or serial port on the PC or notebook below
 * in the source code. Run this app, the default settings should be OK if the
 * Digital One device still has its factory settings with the exception that the MODE
 * SWITCH is in the ON (upper) position.
 * 
 */

namespace PiXtend_eIO_Demo_Console_App
{
    public class SerialConsoleApp
    {
        static bool _continue;
        static SerialPort _serialPort;
        //Counter for the current command to send
        static int _cmdCount;
        //Array of command to send
        static readonly string[] _sCmds =  {"#001:004:01:000*",
                                            "#001:004:01:001*",
                                            "#001:004:01:002*",
                                            "#001:004:01:004*",
                                            "#001:004:01:008*",
                                            "#001:004:01:016*",
                                            "#001:004:01:032*",
                                            "#001:004:01:064*",
                                            "#001:004:01:128*",
                                            "#001:004:01:064*",
                                            "#001:004:01:032*",
                                            "#001:004:01:016*",
                                            "#001:004:01:008*",
                                            "#001:004:01:004*",
                                            "#001:004:01:002*",
                                            "#001:004:01:001*"};

        public static void Main()
        {
            string message;
            StringComparer stringComparer = StringComparer.OrdinalIgnoreCase;
            //Create thread for communication
            Thread readThread = new Thread(DoCommunication);

            // Create a new SerialPort object with default settings.
            _serialPort = new SerialPort();

            // Allow the user to set the appropriate properties.
            _serialPort.PortName = SetPortName("COM9");
            _serialPort.BaudRate = SetPortBaudRate(19200);
            _serialPort.Parity = Parity.Even;
            _serialPort.Parity = SetPortParity(_serialPort.Parity);
            _serialPort.DataBits = SetPortDataBits(_serialPort.DataBits);
            _serialPort.StopBits = SetPortStopBits(_serialPort.StopBits);
            _serialPort.Handshake = SetPortHandshake(_serialPort.Handshake);

            // Set the read/write timeouts
            _serialPort.ReadTimeout = 500;
            _serialPort.WriteTimeout = 500;
            _cmdCount = 0;

            //Open the serial port
            _serialPort.Open();
            _continue = true;
            //Start the thread
            readThread.Start();

            Console.WriteLine("Type QUIT to exit");
            //Wait for the user to enter the word quit
            while (_continue)
            {
                message = Console.ReadLine();

                if (stringComparer.Equals("quit", message))
                {
                    _continue = false;
                }
            }
            //End thread for communication
            readThread.Join();
            //Close the serial port
            _serialPort.Close();
        }

        public static void DoCommunication()
        {
            while (_continue)
            {
                try
                {
                    //Clear the buffer of the serial port 
                    _serialPort.DiscardInBuffer();
                    //Write the selected command from the array to the RS485 bus
                    _serialPort.Write(_sCmds[_cmdCount]);
                    System.Threading.Thread.Sleep(5);
                    //Read back the reply from the Digital One device
                    string message1 = _serialPort.ReadTo("*");
                    //Check if the sent message is the same as the received message minus the * character
                    //The length of one message is 16, but minus 1 for the missing * character so 15
                    if (message1 != _sCmds[_cmdCount].Substring(0,15) || message1.Length != 15)
                    {
                        Console.WriteLine("Error: Reply for setting Outputs not correct!");
                        System.Threading.Thread.Sleep(9900);
                    }
                    //Increment array counter for the next command, if we go over 15, the we start from the beginning
                    _cmdCount++;
                    if (_cmdCount > 15)
                    {
                        _cmdCount = 1;
                    }
                    //Short sleep
                    System.Threading.Thread.Sleep(100);
                }
                catch (TimeoutException) { }
            }
        }

        // Display Port values and prompt user to enter a port.
        public static string SetPortName(string defaultPortName)
        {
            string portName;

            Console.WriteLine("Available Ports:");
            foreach (string s in SerialPort.GetPortNames())
            {
                Console.WriteLine("   {0}", s);
            }

            Console.Write("Enter COM port value (Default: {0}): ", defaultPortName);
            portName = Console.ReadLine();

            if (portName == "" || !(portName.ToLower()).StartsWith("com"))
            {
                portName = defaultPortName;
            }
            return portName;
        }
        // Display BaudRate values and prompt user to enter a value.
        public static int SetPortBaudRate(int defaultPortBaudRate)
        {
            string baudRate;

            Console.Write("Baud Rate(default:{0}): ", defaultPortBaudRate);
            baudRate = Console.ReadLine();

            if (baudRate == "")
            {
                baudRate = defaultPortBaudRate.ToString();
            }

            return int.Parse(baudRate);
        }

        // Display PortParity values and prompt user to enter a value.
        public static Parity SetPortParity(Parity defaultPortParity)
        {
            string parity;

            Console.WriteLine("Available Parity options:");
            foreach (string s in Enum.GetNames(typeof(Parity)))
            {
                Console.WriteLine("   {0}", s);
            }

            Console.Write("Enter Parity value (Default: {0}):", defaultPortParity.ToString(), true);
            parity = Console.ReadLine();

            if (parity == "")
            {
                parity = defaultPortParity.ToString();
            }

            return (Parity)Enum.Parse(typeof(Parity), parity, true);
        }
        // Display DataBits values and prompt user to enter a value.
        public static int SetPortDataBits(int defaultPortDataBits)
        {
            string dataBits;

            Console.Write("Enter DataBits value (Default: {0}): ", defaultPortDataBits);
            dataBits = Console.ReadLine();

            if (dataBits == "")
            {
                dataBits = defaultPortDataBits.ToString();
            }

            return int.Parse(dataBits.ToUpperInvariant());
        }

        // Display StopBits values and prompt user to enter a value.
        public static StopBits SetPortStopBits(StopBits defaultPortStopBits)
        {
            string stopBits;

            Console.WriteLine("Available StopBits options:");
            foreach (string s in Enum.GetNames(typeof(StopBits)))
            {
                Console.WriteLine("   {0}", s);
            }

            Console.Write("Enter StopBits value (None is not supported and \n" +
             "raises an ArgumentOutOfRangeException. \n (Default: {0}):", defaultPortStopBits.ToString());
            stopBits = Console.ReadLine();

            if (stopBits == "")
            {
                stopBits = defaultPortStopBits.ToString();
            }

            return (StopBits)Enum.Parse(typeof(StopBits), stopBits, true);
        }
        public static Handshake SetPortHandshake(Handshake defaultPortHandshake)
        {
            string handshake;

            Console.WriteLine("Available Handshake options:");
            foreach (string s in Enum.GetNames(typeof(Handshake)))
            {
                Console.WriteLine("   {0}", s);
            }

            Console.Write("Enter Handshake value (Default: {0}):", defaultPortHandshake.ToString());
            handshake = Console.ReadLine();

            if (handshake == "")
            {
                handshake = defaultPortHandshake.ToString();
            }

            return (Handshake)Enum.Parse(typeof(Handshake), handshake, true);
        }
    }
}
