#!/usr/bin/python

# This file is part of the PiXtend(R) Project.
#
# For more information about PiXtend(R) and this program,
# see <https://www.pixtend.de> or <https://www.pixtend.com>
#
# Copyright (C) 2019 Robin Turner
# Qube Solutions GmbH, Arbachtalstr. 6
# 72800 Eningen, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Made for Python 3 and pySerial

from __future__ import print_function
import serial, time

#    ------------------------------------
#    Digital One
#    ------------------------------------
#    Protocol length: 16
#    - Set DO0 of Device 37 to High
#      #001:003:01:001*
#    - Set DO0 of Device 12 to Low
#      #001:003:01:000*
#    - Read DI1 of Device 12
#      #001:001:02:002*
#
#    ------------------------------------
#    Analog One
#    ------------------------------------
#    Protocol length: 17
#    - Set AO0 to 
#      #003:006:01:4095*
#    - Set AO0 to
#      #003:006:01:0000*


#Create an array of strings
sCmds = ['' for x in range(16)]
#Assign a command to each element
#Write/Set all Digital Outputs to 0 (off)
sCmds[0] = '#001:004:01:000*'
sCmds[1] = '#001:004:01:001*'
sCmds[2] = '#001:004:01:002*'
sCmds[3] = '#001:004:01:004*'
sCmds[4] = '#001:004:01:008*'
sCmds[5] = '#001:004:01:016*'
sCmds[6] = '#001:004:01:032*'
sCmds[7] = '#001:004:01:064*'
sCmds[8] = '#001:004:01:128*'
sCmds[9] = '#001:004:01:064*'
sCmds[10] = '#001:004:01:032*'
sCmds[11] = '#001:004:01:016*'
sCmds[12] = '#001:004:01:008*'
sCmds[13] = '#001:004:01:004*'
sCmds[14] = '#001:004:01:002*'
sCmds[15] = '#001:004:01:001*'

#Initialization and open the serial port
ser = serial.Serial()
#Set the com port name, Windows: COM1-COM255, Linux: /dev/ttyS0 or /dev/ttyUSB0 etc.
ser.port = 'COM9'
#Choose baudrate, usually we want 115200, but lower speeds are possible, see manual
ser.baudrate = 19200
#Number of bits per byte, we use 8 bits
ser.bytesize = serial.EIGHTBITS
#Set parity, here we use "no parity", but Digital One/Analog One modules also offer "even" or "odd" parity
ser.parity = serial.PARITY_EVEN
#Number of stop bits to use, usually this is set to one stop bit
ser.stopbits = serial.STOPBITS_ONE
#Blocked read, wait for data to arrive or a timeout error will happen
ser.timeout = 2
#Set a timeout for writing, in case the serial port hangs while writing
ser.writeTimeout = 2
#Turn off software flow control         
ser.xonxoff = False
#Turn off hardware (RTS/CTS) flow control   
ser.rtscts = False
#Turn off hardware (DSR/DTR) flow control
ser.dsrdtr = False       

try: 
    #Try to open the com port, so we can start to send commands to the module
    ser.open()
except Exception as e:
    print ("error open serial port: " + str(e))
    exit()

#Could the serial port be opned?
if ser.isOpen():

    try:
        #flush input buffer, discarding all its contents
        ser.flushInput()
        #flush output buffer, aborting current output and discard all that is in the buffer      
        ser.flushOutput()

        #Write the first command to the bus
        ser.write(sCmds[0].encode())
        #Print out the response on the console so we have a feedback at start
        print("Start: " + sCmds[0])
        
        #Short pause before we start the main loop and getting back our first response
        time.sleep(0.5)
        #Read from the serial port until a * char is found, this is the PiXtend eIO Protocol End-Marker
        #Also the total number of bytes is 16, this is the length of a command for a Digital One module
        response = ser.read_until('*', size=16)
        #Print out the response, so we can see it in the console
        print("Response: " + response.decode())        
        
        #Set up some helper variables
        #Current array element containing a command
        cmdcount = 1
        #Was the first command, Write Digital Outputs, successful
        cmd_1_ok = 0
        #Simple loop counter
        cyclecounter = 0

        #Start while loop
        while True:
            #Write command string from array to RS485 bus
            ser.write(sCmds[cmdcount].encode())
            #Read back the response of the module. Read until a * char is found
            #and the length of the response is 16 bytes as per definition for a Digital One module
            response1 = ser.read_until('*', size=16)
            #For demonstration purposes we check only the length here
            if (len(response1.decode()) == 16):
                #Set first helper variable, response to first command is OK
                cmd_1_ok = 1
                    
            #Increment counter for next command string in array variable sCmds
            cmdcount += 1
            #Check the command counter variable if it has already reached the maximum numner of available elements
            if (cmdcount >= 16):
                #Reset command counter to 1, not 0, we want running lights
                #the command at element position 0 would turn off all lights and this couls lead to
                #unwanted bliking of the digital output led 0
                cmdcount = 1
            #Dig we get a good response?
            if (cmd_1_ok == 1):
                #All went OK, so reset the helper variables
                cmd_1_ok = 0
            else:
                #There was an error, the digital output we wanted to turn on is not on
                print("Error communicating with the device!!!! "+str(cyclecounter))
                cmd_1_ok = 0           
                time.sleep(15)
            #Wait 100ms just for this demonstration, if the used computer or Raspberry Pi
            #is fast enough, the sleep command can be deleted for maximum performace.
            #But this is not recommened, there should alway be a small pause.
            time.sleep(0.1)
            #Increment cycle counter
            cyclecounter+=1            
        #When leaning the loop, close the com port
        ser.close()
    except Exception as e1:
        print ("error communicating...: " + str(e1))

else:
    print ("cannot open serial port ")